// observe_ime_wnd.cpp Copyright (c) 2018 okiraku-camera.tokyo
//
// This software is released under the MIT License.
// http://opensource.org/licenses/mit-license.php

#include "stdafx.h"
#include "observe_ime.h"
#include "observe_ime_wnd.h"
#include <imm.h>


static const int observe_interval = 100;

Cobserve_ime_wnd::Cobserve_ime_wnd() {
	m_timer_id = 0;
	m_trayMenu = 0;
	m_trayIcon = 0;
	m_enabled = false;
}

BEGIN_MESSAGE_MAP(Cobserve_ime_wnd, CWnd)
	ON_WM_CREATE()
	ON_WM_TIMER()
	ON_MESSAGE(PM_SHELLNOTIFY, &Cobserve_ime_wnd::OnPmShellnotify)
	ON_MESSAGE(IDM_EXIT_APP, &Cobserve_ime_wnd::OnIdmExitApp)
	ON_WM_POWERBROADCAST()
END_MESSAGE_MAP()

void Cobserve_ime_wnd::PostNcDestroy() {
	CWnd::PostNcDestroy();
	delete this;
}

bool Cobserve_ime_wnd::Create() {
	LPCTSTR lpcClass = AfxRegisterWndClass(CS_VREDRAW | CS_HREDRAW);
	HWND hwnd = ::FindWindow(lpcClass, AfxGetAppName());
	if (hwnd != 0)
		return false;
	return CreateEx(0, lpcClass, AfxGetAppName(), WS_POPUP, CRect(0, 0, 0, 0), NULL , NULL);
}

int Cobserve_ime_wnd::OnCreate(LPCREATESTRUCT lpCreateStruct) {
	if (CWnd::OnCreate(lpCreateStruct) == -1)
		return -1;
	m_trayIcon = theApp.LoadIcon(IDI_ICON2);
	m_timer_id = SetTimer(1, observe_interval, NULL);
	m_enabled = true;
	SetNotifyIcon();
	return 0;
}

void Cobserve_ime_wnd::OnDestroy() {
	CWnd::OnDestroy();
	send_keycode(VK_SCROLL, off);
	if (m_timer_id)
		KillTimer(m_timer_id);
	m_enabled = false;
}

afx_msg LRESULT Cobserve_ime_wnd::OnIdmExitApp(WPARAM wParam, LPARAM lParam) {
	DestroyWindow();
	return 0;
}

void Cobserve_ime_wnd::SetMenuChecked(UINT w) {
	if (m_trayMenu)
		CheckMenuItem(m_trayMenu, w, MF_BYCOMMAND | m_enabled ? MF_CHECKED : MF_UNCHECKED);
}


BOOL Cobserve_ime_wnd::OnCommand(WPARAM wParam, LPARAM lParam) {
	if (wParam == IDM_EXIT_APP) {
		PostMessage(WM_CLOSE);
		DeleteNotifyIcon();
	} else if (wParam == IDM_OBSERVING) {
		if (m_timer_id) {
			KillTimer(m_timer_id);
			m_timer_id = 0;
			m_enabled = false;
		}
		else {
			m_timer_id = SetTimer(1, observe_interval, NULL);
			m_enabled = true;
		}
		SetMenuChecked(IDM_OBSERVING);
	}
	return CWnd::OnCommand(wParam, lParam);
}

void Cobserve_ime_wnd::send_keycode(WORD vkey, key_action_t action) {

	if (action == on || action == off) {
		SHORT u = GetKeyState(vkey);
		if ((u & 1) && action == on)
			return;
		if (!(u & 1) && action == off)
			return;
	}
	INPUT input = { 0 };
	input.type = INPUT_KEYBOARD;
	input.ki.wVk = vkey;
	input.ki.dwFlags = KEYEVENTF_EXTENDEDKEY;
	SendInput(1, &input, sizeof(INPUT));

	input.ki.dwFlags = KEYEVENTF_EXTENDEDKEY | KEYEVENTF_KEYUP;
	SendInput(1, &input, sizeof(INPUT));
}

#ifndef IMC_GETCONVERSIONMODE 
#define IMC_GETCONVERSIONMODE 1
#endif
#ifndef IMC_GETOPENSTATUS 
#define IMC_GETOPENSTATUS 5
#endif

void Cobserve_ime_wnd::OnTimer(UINT_PTR id) {
	CWnd::OnTimer(id);
	if (id == m_timer_id) {
		bool status = false;
		static bool last_status = false;
		GUITHREADINFO gti = { 0 };
		gti.cbSize = sizeof(GUITHREADINFO);
		GetGUIThreadInfo(NULL, &gti);
		HWND hwndime = ImmGetDefaultIMEWnd(gti.hwndFocus);
		if (!IsWindow(hwndime))
			return;
		if (::SendMessage(hwndime, WM_IME_CONTROL, IMC_GETOPENSTATUS, 0) == 1) {
			LRESULT n = ::SendMessage(hwndime, WM_IME_CONTROL, IMC_GETCONVERSIONMODE, 0);
			if (n & 1)
				status = true;
		}
		if (status != last_status) {
			TRACE(_T("%04x\n"), status);
			last_status = status;
			send_keycode(VK_SCROLL, status ? on : off);
		}
	}
}

void Cobserve_ime_wnd::SetNotifyIcon() {

	NOTIFYICONDATA	nic;
	memset(&nic, 0, sizeof(NOTIFYICONDATA));
	nic.cbSize = sizeof(NOTIFYICONDATA);
	nic.hWnd = m_hWnd;
	nic.uID = IDI_ICON2;

	nic.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
	nic.uCallbackMessage = PM_SHELLNOTIFY;
	nic.hIcon = m_trayIcon;
	lstrcpy(nic.szTip, _T("observe_ime"));
	Shell_NotifyIcon(NIM_ADD, &nic);
}

void Cobserve_ime_wnd::DeleteNotifyIcon() {
	NOTIFYICONDATA	nic;
	memset(&nic, 0, sizeof(NOTIFYICONDATA));
	nic.cbSize = sizeof(NOTIFYICONDATA);
	nic.hWnd = m_hWnd;
	nic.uID = IDI_ICON2;
	Shell_NotifyIcon(NIM_DELETE, &nic);
}

afx_msg LRESULT Cobserve_ime_wnd::OnPmShellnotify(WPARAM wParam, LPARAM lParam)
{
	POINT point;
	switch (LOWORD(lParam)) {
	case WM_RBUTTONDOWN:
		if (m_trayMenu == 0) {
			m_trayMenu = CreatePopupMenu();
			if (m_trayMenu) {
				AppendMenu(m_trayMenu, MF_STRING, IDM_OBSERVING, _T("IMEĎ"));
				AppendMenu(m_trayMenu, MF_SEPARATOR, 0, NULL);
				AppendMenu(m_trayMenu, MF_STRING, IDM_EXIT_APP, _T("I"));
			}
		}
		SetMenuChecked(IDM_OBSERVING);
		GetCursorPos(&point);
		SetForegroundWindow();
		TrackPopupMenu(m_trayMenu, TPM_VERTICAL, point.x, point.y, 0, m_hWnd, NULL);
		break;
	default:
		break;
	}
	return 1;
}

UINT Cobserve_ime_wnd::OnPowerBroadcast(UINT nPowerEvent, LPARAM nEventData)
{
	if (nPowerEvent == PBT_APMSUSPEND) {
		KillTimer(m_timer_id);
		m_timer_id = 0;
		send_keycode(VK_SCROLL, off);	// ԂɍĂȂB
	}
	else if (nPowerEvent == PBT_APMRESUMEAUTOMATIC) {
		send_keycode(VK_SCROLL, off);	// ͗L̂悤łB
		if (m_enabled && m_timer_id == 0)
			m_timer_id = SetTimer(1, observe_interval, NULL);
	}

	return CWnd::OnPowerBroadcast(nPowerEvent, nEventData);
}
